/****************************************************************/
/****************************************************************/
/*                                                              */
/*   Anonymizer example console application                     */
/*   for help run:                                              */
/*     anonymizer-cmd --help                                    */
/*                                                              */
/* ------------------------------------------------------------ */
/*                                                              */
/*    Eyedea Recognition, Ltd. (C) 2013, Mar 6th                */
/*                                                              */
/*    Contact:                                                  */
/*               web: http://www.eyedea.cz                      */
/*             email: info@eyedea.cz                            */
/*                                                              */
/****************************************************************/
/****************************************************************/

#include <errno.h>
#include <limits.h>
#include <cstring>
#include <sstream>
#include <fstream>
#include <vector>
#include <algorithm>
#include <iterator>

#include <er_explink.h> /* multi-platfoprm explicit linking macros */
#include <Anonymizer.h> /* Anonymizer API */

#define ANONYMIZER_SDK_DIR "../../AnonymizerSDK/"
#define ANONYMIZER_LIB     ANONYMIZER_SDK_DIR "/lib/" ER_LIB_PREFIX "anonymizer-" ER_LIB_TARGET ER_LIB_EXT
//#define ANONYMIZER_LIB     ANONYMIZER_SDK_DIR "/lib/" ER_LIB_PREFIX "anonymizer.unsecured-" ER_LIB_TARGET ER_LIB_EXT
//#define ANONYMIZER_LIB     ANONYMIZER_SDK_DIR "/lib/" ER_LIB_PREFIX "anonymizer.unsecured-" ER_LIB_TARGET "d" ER_LIB_EXT
#define ANONYMIZER_CONFIG  "config.ini"

#define VERBOSE(command) if (verbose) \
command

struct FileParts
{
    std::string path;
    std::string name;
    std::string ext;
};

int parseMatrixString(std::string str, std::vector<int>& values);
bool check_arg(const char *arg, std::string option, std::string &retv); // argument checking
FileParts fileparts(std::string filename); // divide string 
bool _isdir(const char *path); // check if directory exist
int  _mkdir(const char *dir);  // create directory - recursively
double get_time();             // get machine time

/* ///////////////////////////////////////////////////////////// */
/*                                                               */
/*   This example uses libeyeanonymize to run several detectors  */
/*   over image and anonomize detections - blurr the detections  */
/*                                                               */
/*                                                               */
/*                                                               */
/* ///////////////////////////////////////////////////////////// */
int main (int argc, const char ** argv)
{
    std::string  output        = "anonymized";
    int          quality       = 90; 
    int          verbose       = 0;
    double       face_thr      = 10.;
    double       lp_thr        = 6;
    std::string  input;
    AnROI        roi = { 0,0,UINT_MAX,UINT_MAX };
    int          face_blur_pass = 1;
    int          lp_blur_pass = 1;
    double       face_size_factor = 1.0;
    double       lp_size_factor = 1.0;
    std::string  config         = ANONYMIZER_CONFIG;

    int anonymization_mode = ANONYMIZE_FACE | ANONYMIZE_LP;
    int show = 0;
    int panoramic= 0;
    int keepEXIF= 0;

    /* ////////////////////////////////////////////////////////// */
    /*  Reading input parameters                                  */
    /*                                                            */
    for( int i = 1; i < argc; i++ )
    {
        std::string retv;
        if (check_arg(argv[i], "-o=", retv) || check_arg(argv[i], "--output=", retv))
            output = retv;
        else if (check_arg(argv[i], "--config=", retv))
            config = retv;
        else if (check_arg(argv[i], "-q=", retv) | check_arg(argv[i], "--quality=", retv))
            quality = atoi(retv.c_str());
        else if (check_arg(argv[i], "--face_thr=", retv))
            face_thr = atof(retv.c_str());
        else if (check_arg(argv[i], "--lp_thr=", retv))
            lp_thr = atof(retv.c_str());
        else if (check_arg(argv[i], "--roi=", retv)) {
            std::vector<int> vroi;
            if (parseMatrixString(retv, vroi)!=0) { printf(" Parsing --roi options failed!\n"); return -1; }
            if (vroi.size() != 4) { printf(" --roi options - number of arguments != 4 \n"); return -1; }
            roi.x = vroi[0]; roi.y = vroi[1]; roi.width = (unsigned int) vroi[2]; roi.height = (unsigned int) vroi[3];
        }
        else if (check_arg(argv[i], "--face_num_blur_pass=", retv))
            face_blur_pass = atoi(retv.c_str());
        else if (check_arg(argv[i], "--lp_num_blur_pass=", retv))
            lp_blur_pass = atoi(retv.c_str());
        else if (check_arg(argv[i], "--face_size_factor=", retv))
            face_size_factor = atof(retv.c_str());
        else if (check_arg(argv[i], "--lp_size_factor=", retv))
            lp_size_factor = atof(retv.c_str());
        else if (check_arg(argv[i], "--lp", retv))
            anonymization_mode = ANONYMIZE_LP;
        else if (check_arg(argv[i], "--face", retv))
            anonymization_mode = ANONYMIZE_FACE;
        else if (check_arg(argv[i], "--show", retv))
            show = 1;
        else if (check_arg(argv[i], "--panoramic", retv))
            panoramic = 1;
        else if (check_arg(argv[i], "--keep_exif", retv))
            keepEXIF = 1;
        else if (check_arg(argv[i], "-v", retv) | check_arg(argv[i], "--verbose", retv))
            verbose = 1;
        else if (check_arg(argv[i], "-h", retv) | check_arg(argv[i], "--help", retv)){
            printf("NAME\n"
            "        anonymizer - anonymizes faces and licence plates at images.\n\n"
            "SYNOPSIS\n"
            "        Unix   : ./anonymizer [options] IMAGE_LIST\n"
            "        Windows: anonymizer.exe [options] IMAGE_LIST\n"
            "\n"
            "DESCRIPTION\n"
            "        The anonymizer program anonymizes images listed at IMAGE_LIST text file.\n"
            "        Each of the image is loaded to the program, faces and licence plates are detected and blurred.\n"
            "        Anonymized images are saved under output directory (see -o, --output option)\n\n"
            "INPUT\n"
            "        IMAGE_LIST - text file with list of images (with abolute or relative path). \n"
            "              Each image filename is at separate line. (Unix: ls -1 *.jpg > image_list.txt)\n\n"
            "OPTIONS\n"
            "        -h, --help this help\n"
            "        -o=directory, --output=directory \n"
            "                   sets the output directory for anonymized images (default 'anonymized')\n"
            "        -q=n, --quality=n \n"
            "                   sets jpeg quality n = 1..100 (90 default)\n"
            "        --keep_exif\n"
            "                   Copy EXIF from original JPG file to output file.\n"
            "                   NOTE that EXIF thumbnail is not anonymized and that some nonstndard EXIFs are not preserved.\n"
            "        --lp       run LP anonymization only\n"
            "        --lp_thr=threshold\n"
            "                   sets the threshold on LP detections\n"
            "        --face     run face annonymization only\n"
            "        --face_thr=threshold\n"
            "                   sets the threshold on face detections\n"
            "        --roi=[x,y,width,height] anonymization ROI (region of interest)\n"
            "                   default, [0 0 UINT_MAX UINT_MAX] \n"
            "        --face_num_blur_pass=n\n"
            "                   sets faces blur amount\n"
            "        --lp_num_blur_pass=n\n"
            "                   sets licence plates blur amount\n"
            "        --face_size_factor=r\n"
            "                   sets size of blured area relative to faces detection size\n"
            "        --lp_size_factor=r\n"
            "                   sets size of blured area relative to lps detection size\n"
            "        --show     colorize detections insted of blurring them\n"
            "        --panoramic\n"
            "                   copy left and right edges of panoramatic images\n"
            "                   to be able to detect on edges of 360 degree image\n"
            "                   only active when ROI width is same as image width\n"
            "        -v, --verbose\n"
            "                   more verbose run.\n"
            "\n"
            "EXAMPLE\n"
            "        anonymizer -o=output_dir input_image_list.txt  \n"
            "                    Anonymize faces and LPs at each image from input_image_list.txt.\n"
            "                    Anonymized images are saved under 'output_dir' directory\n"
            "        anonymizer -o=face_emph_dir --show --face input_image_list.txt\n"
            "                    run face detection only and save images with colorized detections.\n"
            "\n"
            "(C) 2013, Eyedea Recognition s.r.o., http://www.eyedea.cz\n"
            "\n");
            return 0;
        }
        else if (check_arg(argv[i], "-", retv))
            printf("WARNING: Unknown option %s\nSee `%s --help' for more information.\n",argv
            [i], argv[0]);
        else
            input.assign( argv[i] );
    }

    if( input.empty() )
    {
        printf("ERROR: Input image list is missing.\nTry `%s --help' for more information.\n", argv[0]);
        return -1;
    }

    //void    *pState;    /* pointer to state */
    ANState state;
    AnParams parameters;
    parameters.ROI = roi;                             /* limits of Region Of Interest  values <0 are ignored (0 is taken) */
    parameters.mode        = anonymization_mode;      /* anonymize faces and licence plates */
    parameters.jpegQuality = quality;                 /* output JPEG file/buffer quality */

    parameters.faceConfidenceThr = face_thr;          /* all face detections with confidence equal or higher are anonymized */
    parameters.faceDetSizeFactor = face_size_factor;  /* factor to enlarge anonymized area with respect to size of detection */
    parameters.faceNumBlurPass   = face_blur_pass;    /* number of blurr passes over anonymized area */

    parameters.lpConfidenceThr = lp_thr;              /* all LP detections with confidence equal or higher are anonymized */
    parameters.lpDetSizeFactor = lp_size_factor;      /* factor to enlarge anonymized area with respect to size of detection */
    parameters.lpNumBlurPass   = lp_blur_pass;        /* number of blurr passes over anonymized area */
    
    parameters.showDetections  = show;                /* if showDetections>0 than the detections in ouput image are highlighted insted of blurred */
    parameters.verbose         = verbose;             /* if verbose>0 than Anonymizer engine prints to the stdout the progress state */
    parameters.panoramic       = panoramic;           /* if panoramic>0 than Anonymizer engine copy left and right part of panoramatic images to detect objects on the edges of 360° images */
    parameters.keepEXIF        = keepEXIF;
    
    VERBOSE(printf("Loading library and functions ...\n"));
    shlib_hnd                 hDll;
    fcn_anInit                fcnAnInit;
    fcn_anFree                fcnAnFree;
    fcn_anAnonymize           fcnAnAnonymize;
    fcn_anVersion             fcnAnVersion;
    fcn_anGetErrorMsg         fcnAnGetErrorMsg;

#ifdef EXPLICIT_LINK
    /* load shared library and functions */
    ER_OPEN_SHLIB(hDll, ANONYMIZER_LIB);
    if (!hDll){
        printf(ANONYMIZER_LIB);
        printf("  library not loaded!\n");
        return -1;
    }

    /* link  AnonymizerSDK functions */
    ER_LOAD_SHFCN(fcnAnInit, fcn_anInit, hDll, "anInit");
    ER_LOAD_SHFCN(fcnAnFree, fcn_anFree, hDll,  "anFree");
    ER_LOAD_SHFCN(fcnAnAnonymize, fcn_anAnonymize, hDll, "anAnonymize");
    ER_LOAD_SHFCN(fcnAnVersion, fcn_anVersion, hDll, "anVersion");
    ER_LOAD_SHFCN(fcnAnGetErrorMsg, fcn_anGetErrorMsg, hDll, "anGetErrorMsg");

    if (!fcnAnInit || !fcnAnFree || !fcnAnAnonymize || !fcnAnGetErrorMsg || !fcnAnVersion) {
        printf("  Loading Anonymizer functions failed!\n");
        return -1;
    }
#else
    fcnAnInit = &(anInit);
    fcnAnFree = &(anFree);
    fcnAnAnonymize = &(anAnonymize);
    fcnAnVersion   = &(anVersion);
    fcnAnGetErrorMsg = &(anGetErrorMsg);
#endif    
    
    VERBOSE(printf("  ... done\n\n"));

    VERBOSE(printf("Loaded %s\n\n", fcnAnVersion(0)));

    int ern;
    
    /* init EyeAnonymize */
    VERBOSE(printf("Anonymizer initialization ...\n"));
    if ( (ern=fcnAnInit((char*)ANONYMIZER_SDK_DIR,config.c_str(),&state))!=0 ){
        VERBOSE(printf("failed.\n%s\n",fcnAnGetErrorMsg(ern)));
        return -1;
    }else
        VERBOSE(printf("  ... done.\n"));

    if (!_isdir(output.c_str()))
        if (_mkdir(output.c_str())!=0){
            printf("  Could not create '%s' output directory!\n",output.c_str());
            return -1;
        }
            
  
    double start = get_time();
    
    std::ifstream infile(input.c_str());
    std::string line;
    if (!infile.is_open()) printf("Error while opening file %s\n ", input.c_str()); // check if file exist and was successfully opened
    while (std::getline(infile, line))
    {
        if (!line.empty() && line[line.size() - 1] == '\r') line.erase(line.size() - 1);
        if (!line.empty())
        {
            FileParts fp = fileparts(line);
            std::string output_file = output +  "/" + fp.name + ".jpg";
            if (fcnAnAnonymize(line.c_str(),output_file.c_str(),&parameters,state)!=0)
            {
                VERBOSE(printf(" ... FAILED\n%s\n",fcnAnGetErrorMsg(ern)));
            }
            else VERBOSE(printf(" ... done\n"));
        }
    }
    infile.close();

    double end = get_time();
    VERBOSE(printf("Annonymization of images done in %f sec.\n",end-start));
    
    /* free EyeAnonymize state */
    VERBOSE(printf("Cleaning up\n"));
    fcnAnFree(state);

#ifdef EXPLICIT_LINK
     ER_FREE_LIB(hDll); // NOTE: unloading secured library may cause application crash (Windows, randomly)
#endif
    
    VERBOSE(printf("  ... done.\n"));
    return 0;
}

bool check_arg(const char *arg, std::string option, std::string &retv)
{
    if (option.compare(0,option.length(),arg,option.length()) == 0)
    {
        retv.assign(arg+option.length());
        return true;
    }
    else
    {
        return false;
    }
}

static int fixpath(const char *inpath, char *outpath)
{
    int   n=0;
    strcpy(outpath,inpath);
    while(inpath[n]) n++;
#if _WIN32 || _WIN64 
    if(inpath[n-1] != '\\')
        strcat(outpath,"\\");
#else
    if(inpath[n-1] != '/')
        strcat(outpath,"/");
#endif
    return 0;
}


bool _isdir(const char *path) 
{
    /* isdir - tries to open directory (dirent or FindFirsFile), if failed directory doesn't exists */
#if _WIN32 || _WIN64 
    if ((GetFileAttributes(path)) == INVALID_FILE_ATTRIBUTES)
        return false;
    else
        return true;
#else
    char path_[260];
    DIR *dp;
    fixpath(path,path_); /* add / at the end if not present */
    dp = opendir(path_);
    if (dp == NULL)
        return 0;
    closedir(dp);
#endif
    return 1;
}


int _mkdir(const char *dir)
{
    char dir_[260];
    fixpath(dir,dir_); 
#if _WIN32 || _WIN64 
    DWORD ret = 0;
    size_t n;

    char * p = strrchr((char*)dir_, '\\');
    char * q = strrchr((char*)dir_, '/'); // allow unix path separator

    if (p < q)
        p = q;

    if (p==NULL)
        return -1;

    // try to create
    if (CreateDirectory(dir_,0))
        return 0;
    else
    {
        ret = GetLastError();
        if (ret == ERROR_ALREADY_EXISTS)
            return 0;
    }

    // not created - do it recursively
    n   = p-dir_;
    p   =strncpy((char *)malloc(1+n), dir_, n);
    p[n]='\0';
    ret = _mkdir(p);
    free(p);

    if (!ret)
    {
        if (CreateDirectory(dir_,0))
            return 0;
        else
        {
            ret = GetLastError();
            if (ret == ERROR_ALREADY_EXISTS)
                return 0;
        }
    }
    return (int)ret;

#else
    char * p = strrchr((char*)dir_, '/'); 

    int ret = 0;
    size_t n;

    if (p==NULL)
        return -1;

    // try to create
    if (mkdir(dir_,S_IRWXU) == 0)
        return 0;
    else
    {
        if (errno == EEXIST)
            return 0;
    }

    // not created - do it recursively
    n   = p-dir_;
    p   =strncpy((char *)malloc(1+n), dir_, n); p[n]='\0';
    ret = _mkdir(p);
    free(p);

    if (!ret)
    {
        if (mkdir(dir_,S_IRWXU) == 0)
            return 0;
        else
        {
            if (errno == EEXIST)
                return 0;
        }
    }
    return (int)ret;
#endif
}

double get_time()
{
#if _WIN32 || _WIN64
    return GetTickCount()/1000.;
#elif __linux__ || __unix
    struct timespec ts;
    if (!clock_gettime(CLOCK_REALTIME, &ts))
        return (double)((1000*1000*1000) * ts.tv_sec + ts.tv_nsec) / 1000000000.;
    else
        return 0.;
#else
#   error Unsupported OS!
#endif
}

FileParts fileparts(std::string filename) // NOTE: this function does not handle extensions like tar.gz correctly
{
    size_t idx0  = filename.rfind("/");
    size_t idx0_ = filename.rfind("\\");
    size_t idx1 = filename.rfind(".");
    
    if (idx0_ != std::string::npos && (idx0 == std::string::npos || idx0_>idx0 ) ) idx0= idx0_; // select higher value from idx0 and idx0_ but be aware of string::npos value
    if (idx1<idx0 && idx0!= std::string::npos) idx1=filename.length(); // filename doesn't have extension
    if (idx1==idx0+1) idx1=filename.length(); // filename does not have name part (e.g. ".file")
    if (idx1== std::string::npos) idx1=filename.length(); // filename does not have name ext part (e.g. "file")
    
    FileParts fp;
    fp.path = filename.substr(0,idx0+1);
    fp.name = filename.substr(idx0+1,idx1-idx0-1);
    fp.ext  = filename.substr(idx1);

    return fp;
}

int parseMatrixString(std::string str, std::vector<int>& values)
{
    values.clear();

    if (str == "" || str == "[]")
    {
        return 0;
    }

    try
    {
        std::replace(str.begin(), str.end(), '[', ' ');
        std::replace(str.begin(), str.end(), ']', ' ');
        std::replace(str.begin(), str.end(), ',', ' ');
        std::replace(str.begin(), str.end(), ';', ' ');

        std::istringstream iss(str);
        std::vector<std::string> tokens{ std::istream_iterator<std::string>{iss}, std::istream_iterator<std::string>{} };

        values.resize(tokens.size());
        std::transform(tokens.begin(), tokens.end(), values.begin(), [](const std::string& val) {return (int)stod(val); });
    }
    catch (std::bad_alloc e) { printf("%s\n",e.what()); values.clear(); return -1; }
    catch (...) { printf("parsing '[...]' array failed!"); values.clear();  return -1; }
    return 0;
}
